<?php

namespace Drupal\entity_usage_explorer;

use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\ContentEntityInterface;

/**
 * Provides a helper functions for module.
 */
class UsageService {

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Constructs a new UsageService object.
   *
   * @param \Drupal\Core\Database\Connection $database
   *   The database connection.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(Connection $database, EntityTypeManagerInterface $entity_type_manager) {
    $this->database = $database;
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * Summary of loadContentEntityTypes.
   */
  public function loadContentEntityTypes() : array {
    $entity_definitions = $this->entityTypeManager->getDefinitions();

    $content_entity_types = [];

    foreach ($entity_definitions as $entity_type_id => $definition) {
      if ($definition->entityClassImplements(ContentEntityInterface::class)) {
        $content_entity_types[] = $entity_type_id;
      }
    }

    return $content_entity_types;
  }

  /**
   * Implements getEntityReferences().
   */
  public function getEntityReferences($entity_type, $bundle_to_query) {
    $target_type = '';
    $entity_tables = $this->database->query("SHOW TABLES LIKE '%{$entity_type}__%'")->fetchAllKeyed(0, 0);
    $matches = [];

    foreach ($entity_tables as $table) {
      $prefix_key = str_replace("{$entity_type}__", '', $table);
      $column = $this->database->query("SHOW COLUMNS FROM `$table` LIKE '%target_id%'")->fetch();

      if ($column) {
        $column_name = $column->Field;
        if (str_ends_with($column_name, '_target_id')) {
          $suffix_key = str_replace('_target_id', '', $column_name);

          if ($prefix_key === $suffix_key) {
            $field_name = $prefix_key;
            $storage = $this->entityTypeManager->getStorage('field_storage_config')->load("{$entity_type}." . $field_name);
            if ($storage && in_array($storage->getType(), ['entity_reference', 'entity_reference_revisions', 'file'])) {
              $target_type = $storage->getSettings()['target_type'];
            }
            if ($bundle_to_query == $target_type) {
              $matches[] = [
                'table' => $table,
                'column' => $column_name,
                'field_name' => $field_name,
                'key' => $prefix_key,
                'target_type' => $target_type ?? 'Not an entity reference field',
              ];
            }
          }
        }
      }
    }

    return $matches;
  }

  /**
   * Implements getEntityUsage().
   */
  public function getEntityUsage(string $target_entity, int $entity_id): array {
    $data = [];
    $entity_types = $this->loadContentEntityTypes();
    foreach ($entity_types as $entity_type) {
      if ($entity_type == 'menu_link_content') {
        $menu_record = $this->getEntityUsageInMenu($target_entity, $entity_id);
        if (!empty($menu_record)) {
          $data[$entity_type] = $menu_record;
        }
      }

      $matches = $this->getEntityReferences($entity_type, $target_entity);
      if (!empty($matches)) {
        $data[$entity_type] = [];
        foreach ($matches as $match) {
          $query = $this->database->select($match['table'], 't')
            ->fields('t')
            ->condition($match['column'], $entity_id)
            ->execute();

          $result = $query->fetchAll();
          if (!empty($result)) {
            $data[$entity_type] = array_merge($data[$entity_type], $result);
          }
        }
      }
    }

    return $data;
  }

  /**
   * Implements getEntityUsageCount().
   */
  public function getEntityUsageCount(string $target_entity, int $entity_id): int {
    $count = 0;
    $entity_types = $this->loadContentEntityTypes();
    foreach ($entity_types as $entity_type) {
      $matches = $this->getEntityReferences($entity_type, $target_entity);
      if (!empty($matches)) {
        foreach ($matches as $match) {
          $query = $this->database->select($match['table'], 't')
            ->condition($match['column'], $entity_id)
            ->countQuery()
            ->execute();
          $result = $query->fetchField();
          $count += !empty($result) ? $result : 0;
        }
      }
    }
    $count += !empty($this->getEntityUsageInMenu($target_entity, $entity_id, TRUE)) ? $this->getEntityUsageInMenu($target_entity, $entity_id, TRUE) : 0;

    return $count;
  }

  /**
   * Implements getEntityUsageInMenu().
   */
  public function getEntityUsageInMenu(string $target_entity, int $entity_id, $only_count = FALSE) {
    $query = $this->database->select('menu_link_content_data', 'm');
    $group = $query->orConditionGroup()
      ->condition('link__uri', "entity:{$target_entity}/{$entity_id}", 'LIKE')
      ->condition('link__uri', "internal:/{$target_entity}/{$entity_id}", 'LIKE');

    $result = $only_count ? $query->condition($group)->countQuery()->execute()->fetchField() : $query->fields('m')->condition($group)->execute()->fetchAll();

    return !empty($result) ? $result : NULL;
  }

}
